@@ -46,7 +46,7 @@ with open(join(root, '.mailmap')) as f:
         mailmap[src] = dest
         email_names[dest] = line[: line.index('<')].strip()
 
-author_commits = defaultdict(lambda: [])
+author_commits = defaultdict(list)
 
 for commit in get_all_commits():
     # exclude some specific commits (e.g. docstring typos)
@@ -82,15 +82,5 @@ for email, commits in sorted(author_commits.items(), key=sort_key, reverse=True)
             commits[0].authored_datetime.year,
         )
     else:
-        msg = "{commits} commits ({start}-{end})".format(
-            commits=len(commits),
-            start=commits[-1].authored_datetime.year,
-            end=commits[0].authored_datetime.year,
-        )
-    print(
-        "- [ ] {name} {email}: {msg}".format(
-            name=email_names[email],
-            email=email,
-            msg=msg,
-        )
-    )
+        msg = f"{len(commits)} commits ({commits[-1].authored_datetime.year}-{commits[0].authored_datetime.year})"
+    print(f"- [ ] {email_names[email]} {email}: {msg}")
