@@ -2,36 +2,30 @@
 
 # Copyright (c) PyZMQ Developers.
 # Distributed under the terms of the Modified BSD License.
+from __future__ import annotations
 
 import warnings
 from asyncio import Future
 from collections import deque
+from functools import partial
 from itertools import chain
 from typing import (
     Any,
     Awaitable,
     Callable,
-    Dict,
-    List,
     NamedTuple,
-    Optional,
-    Tuple,
-    Type,
     TypeVar,
-    Union,
     cast,
-    overload,
 )
 
 import zmq as _zmq
 from zmq import EVENTS, POLLIN, POLLOUT
-from zmq._typing import Literal
 
 
 class _FutureEvent(NamedTuple):
     future: Future
     kind: str
-    kwargs: Dict
+    kwargs: dict
     msg: Any
     timer: Any
 
@@ -49,7 +43,7 @@ class _Async:
     """Mixin for common async logic"""
 
     _current_loop: Any = None
-    _Future: Type[Future]
+    _Future: type[Future]
 
     def _get_loop(self) -> Any:
         """Get event loop
@@ -78,10 +72,10 @@ class _Async:
 class _AsyncPoller(_Async, _zmq.Poller):
     """Poller that returns a Future on poll, instead of blocking."""
 
-    _socket_class: Type["_AsyncSocket"]
+    _socket_class: type[_AsyncSocket]
     _READ: int
     _WRITE: int
-    raw_sockets: List[Any]
+    raw_sockets: list[Any]
 
     def _watch_raw_socket(self, loop: Any, socket: Any, evt: int, f: Callable) -> None:
         """Schedule callback for a raw socket"""
@@ -91,7 +85,7 @@ class _AsyncPoller(_Async, _zmq.Poller):
         """Unschedule callback for a raw socket"""
         raise NotImplementedError()
 
-    def poll(self, timeout=-1) -> Awaitable[List[Tuple[Any, int]]]:  # type: ignore
+    def poll(self, timeout=-1) -> Awaitable[list[tuple[Any, int]]]:  # type: ignore
         """Return a Future for a poll event"""
         future = self._Future()
         if timeout == 0:
@@ -109,7 +103,7 @@ class _AsyncPoller(_Async, _zmq.Poller):
         watcher = self._Future()
 
         # watch raw sockets:
-        raw_sockets: List[Any] = []
+        raw_sockets: list[Any] = []
 
         def wake_raw(*args):
             if not watcher.done():
@@ -119,11 +113,18 @@ class _AsyncPoller(_Async, _zmq.Poller):
             lambda f: self._unwatch_raw_sockets(loop, *raw_sockets)
         )
 
+        wrapped_sockets: list[_AsyncSocket] = []
+
+        def _clear_wrapper_io(f):
+            for s in wrapped_sockets:
+                s._clear_io_state()
+
         for socket, mask in self.sockets:
             if isinstance(socket, _zmq.Socket):
                 if not isinstance(socket, self._socket_class):
                     # it's a blocking zmq.Socket, wrap it in async
                     socket = self._socket_class.from_socket(socket)
+                    wrapped_sockets.append(socket)
                 if mask & _zmq.POLLIN:
                     socket._add_recv_event('poll', future=watcher)
                 if mask & _zmq.POLLOUT:
@@ -159,6 +160,9 @@ class _AsyncPoller(_Async, _zmq.Poller):
 
         watcher.add_done_callback(on_poll_ready)
 
+        if wrapped_sockets:
+            watcher.add_done_callback(_clear_wrapper_io)
+
         if timeout is not None and timeout > 0:
             # schedule cancel to fire on poll timeout, if any
             def trigger_timeout():
@@ -194,13 +198,12 @@ T = TypeVar("T", bound="_AsyncSocket")
 
 
 class _AsyncSocket(_Async, _zmq.Socket[Future]):
-
     # Warning : these class variables are only here to allow to call super().__setattr__.
     # They be overridden at instance initialization and not shared in the whole class
     _recv_futures = None
     _send_futures = None
     _state = 0
-    _shadow_sock: "_zmq.Socket"
+    _shadow_sock: _zmq.Socket
     _poller_class = _AsyncPoller
     _fd = None
 
@@ -209,7 +212,7 @@ class _AsyncSocket(_Async, _zmq.Socket[Future]):
         context=None,
         socket_type=-1,
         io_loop=None,
-        _from_socket: Optional["_zmq.Socket"] = None,
+        _from_socket: _zmq.Socket | None = None,
         **kwargs,
     ) -> None:
         if isinstance(context, _zmq.Socket):
@@ -234,13 +237,13 @@ class _AsyncSocket(_Async, _zmq.Socket[Future]):
         self._fd = self._shadow_sock.FD
 
     @classmethod
-    def from_socket(cls: Type[T], socket: "_zmq.Socket", io_loop: Any = None) -> T:
+    def from_socket(cls: type[T], socket: _zmq.Socket, io_loop: Any = None) -> T:
         """Create an async socket from an existing Socket"""
         return cls(_from_socket=socket, io_loop=io_loop)
 
-    def close(self, linger: Optional[int] = None) -> None:
+    def close(self, linger: int | None = None) -> None:
         if not self.closed and self._fd is not None:
-            event_list: List[_FutureEvent] = list(
+            event_list: list[_FutureEvent] = list(
                 chain(self._recv_futures or [], self._send_futures or [])
             )
             for event in event_list:
@@ -263,33 +266,9 @@ class _AsyncSocket(_Async, _zmq.Socket[Future]):
 
     get.__doc__ = _zmq.Socket.get.__doc__
 
-    @overload  # type: ignore
-    def recv_multipart(
-        self, flags: int = 0, *, track: bool = False
-    ) -> Awaitable[List[bytes]]:
-        ...
-
-    @overload
-    def recv_multipart(
-        self, flags: int = 0, *, copy: Literal[True], track: bool = False
-    ) -> Awaitable[List[bytes]]:
-        ...
-
-    @overload
-    def recv_multipart(
-        self, flags: int = 0, *, copy: Literal[False], track: bool = False
-    ) -> Awaitable[List[_zmq.Frame]]:  # type: ignore
-        ...
-
-    @overload
     def recv_multipart(
         self, flags: int = 0, copy: bool = True, track: bool = False
-    ) -> Awaitable[Union[List[bytes], List[_zmq.Frame]]]:
-        ...
-
-    def recv_multipart(
-        self, flags: int = 0, copy: bool = True, track: bool = False
-    ) -> Awaitable[Union[List[bytes], List[_zmq.Frame]]]:
+    ) -> Awaitable[list[bytes] | list[_zmq.Frame]]:
         """Receive a complete multipart zmq message.
 
         Returns a Future whose result will be a multipart message.
@@ -300,7 +279,7 @@ class _AsyncSocket(_Async, _zmq.Socket[Future]):
 
     def recv(  # type: ignore
         self, flags: int = 0, copy: bool = True, track: bool = False
-    ) -> Awaitable[Union[bytes, _zmq.Frame]]:
+    ) -> Awaitable[bytes | _zmq.Frame]:
         """Receive a single zmq frame.
 
         Returns a Future, whose result will be the received frame.
@@ -311,7 +290,7 @@ class _AsyncSocket(_Async, _zmq.Socket[Future]):
 
     def send_multipart(  # type: ignore
         self, msg_parts: Any, flags: int = 0, copy: bool = True, track=False, **kwargs
-    ) -> Awaitable[Optional[_zmq.MessageTracker]]:
+    ) -> Awaitable[_zmq.MessageTracker | None]:
         """Send a complete multipart zmq message.
 
         Returns a Future that resolves when sending is complete.
@@ -328,7 +307,7 @@ class _AsyncSocket(_Async, _zmq.Socket[Future]):
         copy: bool = True,
         track: bool = False,
         **kwargs: Any,
-    ) -> Awaitable[Optional[_zmq.MessageTracker]]:
+    ) -> Awaitable[_zmq.MessageTracker | None]:
         """Send a single zmq frame.
 
         Returns a Future that resolves when sending is complete.
@@ -348,6 +327,16 @@ class _AsyncSocket(_Async, _zmq.Socket[Future]):
         def _chain(_):
             """Chain result through serialization to recvd"""
             if f.done():
+                # chained future may be cancelled, which means nobody is going to get this result
+                # if it's an error, that's no big deal (probably zmq.Again),
+                # but if it's a successful recv, this is a dropped message!
+                if not recvd.cancelled() and recvd.exception() is None:
+                    warnings.warn(
+                        # is there a useful stacklevel?
+                        # ideally, it would point to where `f.cancel()` was called
+                        f"Future {f} completed while awaiting {recvd}. A message has been dropped!",
+                        RuntimeWarning,
+                    )
                 return
             if recvd.exception():
                 f.set_exception(recvd.exception())
@@ -384,14 +373,14 @@ class _AsyncSocket(_Async, _zmq.Socket[Future]):
 
         p = self._poller_class()
         p.register(self, flags)
-        f = cast(Future, p.poll(timeout))
+        poll_future = cast(Future, p.poll(timeout))
 
         future = self._Future()
 
         def unwrap_result(f):
             if future.done():
                 return
-            if f.cancelled():
+            if poll_future.cancelled():
                 try:
                     future.cancel()
                 except RuntimeError:
@@ -399,25 +388,30 @@ class _AsyncSocket(_Async, _zmq.Socket[Future]):
                     pass
                 return
             if f.exception():
-                future.set_exception(f.exception())
+                future.set_exception(poll_future.exception())
             else:
-                evts = dict(f.result())
+                evts = dict(poll_future.result())
                 future.set_result(evts.get(self, 0))
 
-        if f.done():
-            # hook up result if
-            unwrap_result(f)
+        if poll_future.done():
+            # hook up result if already done
+            unwrap_result(poll_future)
         else:
-            f.add_done_callback(unwrap_result)
+            poll_future.add_done_callback(unwrap_result)
+
+        def cancel_poll(future):
+            """Cancel underlying poll if request has been cancelled"""
+            if not poll_future.done():
+                try:
+                    poll_future.cancel()
+                except RuntimeError:
+                    # RuntimeError may be called during teardown
+                    pass
+
+        future.add_done_callback(cancel_poll)
+
         return future
 
-    # overrides only necessary for updated types
-    def recv_string(self, *args, **kwargs) -> Awaitable[str]:  # type: ignore
-        return super().recv_string(*args, **kwargs)  # type: ignore
-
-    def send_string(self, s: str, flags: int = 0, encoding: str = 'utf-8') -> Awaitable[None]:  # type: ignore
-        return super().send_string(s, flags=flags, encoding=encoding)  # type: ignore
-
     def _add_timeout(self, future, timeout):
         """Add a timeout for a send or recv Future"""
 
@@ -442,20 +436,22 @@ class _AsyncSocket(_Async, _zmq.Socket[Future]):
         return self._get_loop().call_later(delay, callback)
 
     @staticmethod
-    def _remove_finished_future(future, event_list):
+    def _remove_finished_future(future, event_list, event=None):
         """Make sure that futures are removed from the event list when they resolve
 
         Avoids delaying cleanup until the next send/recv event,
         which may never come.
         """
-        for f_idx, event in enumerate(event_list):
-            if event.future is future:
-                break
-        else:
-            return
-
         # "future" instance is shared between sockets, but each socket has its own event list.
-        event_list.remove(event_list[f_idx])
+        if not event_list:
+            return
+        # only unconsumed events (e.g. cancelled calls)
+        # will be present when this happens
+        try:
+            event_list.remove(event)
+        except ValueError:
+            # usually this will have been removed by being consumed
+            return
 
     def _add_recv_event(self, kind, kwargs=None, future=None):
         """Add a recv event, returning the corresponding Future"""
@@ -479,17 +475,23 @@ class _AsyncSocket(_Async, _zmq.Socket[Future]):
 
         # we add it to the list of futures before we add the timeout as the
         # timeout will remove the future from recv_futures to avoid leaks
-        self._recv_futures.append(_FutureEvent(f, kind, kwargs, msg=None, timer=timer))
-
-        # Don't let the Future sit in _recv_events after it's done
-        f.add_done_callback(
-            lambda f: self._remove_finished_future(f, self._recv_futures)
-        )
+        _future_event = _FutureEvent(f, kind, kwargs, msg=None, timer=timer)
+        self._recv_futures.append(_future_event)
 
         if self._shadow_sock.get(EVENTS) & POLLIN:
             # recv immediately, if we can
             self._handle_recv()
-        if self._recv_futures:
+        if self._recv_futures and _future_event in self._recv_futures:
+            # Don't let the Future sit in _recv_events after it's done
+            # no need to register this if we've already been handled
+            # (i.e. immediately-resolved recv)
+            f.add_done_callback(
+                partial(
+                    self._remove_finished_future,
+                    event_list=self._recv_futures,
+                    event=_future_event,
+                )
+            )
             self._add_io_state(POLLIN)
         return f
 
@@ -538,12 +540,15 @@ class _AsyncSocket(_Async, _zmq.Socket[Future]):
 
         # we add it to the list of futures before we add the timeout as the
         # timeout will remove the future from recv_futures to avoid leaks
-        self._send_futures.append(
-            _FutureEvent(f, kind, kwargs=kwargs, msg=msg, timer=timer)
-        )
+        _future_event = _FutureEvent(f, kind, kwargs=kwargs, msg=msg, timer=timer)
+        self._send_futures.append(_future_event)
         # Don't let the Future sit in _send_futures after it's done
         f.add_done_callback(
-            lambda f: self._remove_finished_future(f, self._send_futures)
+            partial(
+                self._remove_finished_future,
+                event_list=self._send_futures,
+                event=_future_event,
+            )
         )
 
         self._add_io_state(POLLOUT)
@@ -580,7 +585,7 @@ class _AsyncSocket(_Async, _zmq.Socket[Future]):
         elif kind == 'recv':
             recv = self._shadow_sock.recv
         else:
-            raise ValueError("Unhandled recv event type: %r" % kind)
+            raise ValueError(f"Unhandled recv event type: {kind!r}")
 
         kwargs['flags'] |= _zmq.DONTWAIT
         try:
@@ -620,7 +625,7 @@ class _AsyncSocket(_Async, _zmq.Socket[Future]):
         elif kind == 'send':
             send = self._shadow_sock.send
         else:
-            raise ValueError("Unhandled send event type: %r" % kind)
+            raise ValueError(f"Unhandled send event type: {kind!r}")
 
         kwargs['flags'] |= _zmq.DONTWAIT
         try:
@@ -633,6 +638,9 @@ class _AsyncSocket(_Async, _zmq.Socket[Future]):
     # event masking from ZMQStream
     def _handle_events(self, fd=0, events=0):
         """Dispatch IO events to _handle_recv, etc."""
+        if self._shadow_sock.closed:
+            return
+
         zmq_events = self._shadow_sock.get(EVENTS)
         if zmq_events & _zmq.POLLIN:
             self._handle_recv()
