@@ -3,6 +3,7 @@
 # Copyright (C) PyZMQ Developers
 # Distributed under the terms of the Modified BSD License.
 
+from __future__ import annotations
 
 import errno
 import pickle
@@ -11,12 +12,9 @@ import sys
 from typing import (
     Any,
     Callable,
-    Dict,
     Generic,
     List,
-    Optional,
     Sequence,
-    Type,
     TypeVar,
     Union,
     cast,
@@ -25,7 +23,7 @@ from typing import (
 from warnings import warn
 
 import zmq
-from zmq._typing import Literal
+from zmq._typing import Literal, TypeAlias
 from zmq.backend import Socket as SocketBase
 from zmq.error import ZMQBindError, ZMQError
 from zmq.utils import jsonapi
@@ -40,26 +38,28 @@ try:
 except AttributeError:
     DEFAULT_PROTOCOL = pickle.HIGHEST_PROTOCOL
 
-T = TypeVar("T", bound="Socket")
+_SocketType = TypeVar("_SocketType", bound="Socket")
 
 
-class _SocketContext(Generic[T]):
+class _SocketContext(Generic[_SocketType]):
     """Context Manager for socket bind/unbind"""
 
-    socket: T
+    socket: _SocketType
     kind: str
     addr: str
 
     def __repr__(self):
         return f"<SocketContext({self.kind}={self.addr!r})>"
 
-    def __init__(self: "_SocketContext[T]", socket: T, kind: str, addr: str):
+    def __init__(
+        self: _SocketContext[_SocketType], socket: _SocketType, kind: str, addr: str
+    ):
         assert kind in {"bind", "connect"}
         self.socket = socket
         self.kind = kind
         self.addr = addr
 
-    def __enter__(self: "_SocketContext[T]") -> T:
+    def __enter__(self: _SocketContext[_SocketType]) -> _SocketType:
         return self.socket
 
     def __exit__(self, *args):
@@ -71,10 +71,10 @@ class _SocketContext(Generic[T]):
             self.socket.disconnect(self.addr)
 
 
-ST = TypeVar("ST")
+SocketReturnType = TypeVar("SocketReturnType")
 
 
-class Socket(SocketBase, AttributeSetter, Generic[ST]):
+class Socket(SocketBase, AttributeSetter, Generic[SocketReturnType]):
     """The ZMQ socket object
 
     To create a Socket, first create a Context::
@@ -104,38 +104,36 @@ class Socket(SocketBase, AttributeSetter, Generic[ST]):
 
     @overload
     def __init__(
-        self: "Socket[bytes]",
-        ctx_or_socket: "zmq.Context",
+        self: Socket[bytes],
+        ctx_or_socket: zmq.Context,
         socket_type: int,
         *,
-        copy_threshold: Optional[int] = None,
-    ):
-        ...
+        copy_threshold: int | None = None,
+    ): ...
 
     @overload
     def __init__(
-        self: "Socket[bytes]",
+        self: Socket[bytes],
         *,
-        shadow: Union["Socket", int],
-        copy_threshold: Optional[int] = None,
-    ):
-        ...
+        shadow: Socket | int,
+        copy_threshold: int | None = None,
+    ): ...
 
     @overload
     def __init__(
-        self: "Socket[bytes]",
-        ctx_or_socket: "Socket",
-    ):
-        ...
+        self: Socket[bytes],
+        ctx_or_socket: Socket,
+    ): ...
 
     def __init__(
-        self: "Socket[bytes]",
-        ctx_or_socket: Optional[Union["zmq.Context", "Socket"]] = None,
+        self: Socket[bytes],
+        ctx_or_socket: zmq.Context | Socket | None = None,
         socket_type: int = 0,
         *,
-        shadow: Union["Socket", int] = 0,
-        copy_threshold: Optional[int] = None,
+        shadow: Socket | int = 0,
+        copy_threshold: int | None = None,
     ):
+        shadow_context: zmq.Context | None = None
         if isinstance(ctx_or_socket, zmq.Socket):
             # positional Socket(other_socket)
             shadow = ctx_or_socket
@@ -148,6 +146,8 @@ class Socket(SocketBase, AttributeSetter, Generic[ST]):
             # hold a reference to the shadow object
             self._shadow_obj = shadow
             if not isinstance(shadow, int):
+                if isinstance(shadow, zmq.Socket):
+                    shadow_context = shadow.context
                 try:
                     shadow = cast(int, shadow.underlying)
                 except AttributeError:
@@ -162,6 +162,9 @@ class Socket(SocketBase, AttributeSetter, Generic[ST]):
             shadow=shadow_address,
             copy_threshold=copy_threshold,
         )
+        if self._shadow_obj and shadow_context:
+            # keep self.context reference if shadowing a Socket object
+            self.context = shadow_context
 
         try:
             socket_type = cast(int, self.get(zmq.TYPE))
@@ -201,7 +204,7 @@ class Socket(SocketBase, AttributeSetter, Generic[ST]):
         return f"<{_repr_cls}(zmq.{self._type_name}) at {hex(id(self))}{closed}>"
 
     # socket as context manager:
-    def __enter__(self: T) -> T:
+    def __enter__(self: _SocketType) -> _SocketType:
         """Sockets are context managers
 
         .. versionadded:: 14.4
@@ -215,14 +218,14 @@ class Socket(SocketBase, AttributeSetter, Generic[ST]):
     # Socket creation
     # -------------------------------------------------------------------------
 
-    def __copy__(self: T, memo=None) -> T:
+    def __copy__(self: _SocketType, memo=None) -> _SocketType:
         """Copying a Socket creates a shadow copy"""
         return self.__class__.shadow(self.underlying)
 
     __deepcopy__ = __copy__
 
     @classmethod
-    def shadow(cls: Type[T], address: Union[int, "zmq.Socket"]) -> T:
+    def shadow(cls: type[_SocketType], address: int | zmq.Socket) -> _SocketType:
         """Shadow an existing libzmq socket
 
         address is a zmq.Socket or an integer (or FFI pointer)
@@ -261,21 +264,28 @@ class Socket(SocketBase, AttributeSetter, Generic[ST]):
     # Connect/Bind context managers
     # -------------------------------------------------------------------------
 
-    def _connect_cm(self: T, addr: str) -> _SocketContext[T]:
+    def _connect_cm(self: _SocketType, addr: str) -> _SocketContext[_SocketType]:
         """Context manager to disconnect on exit
 
         .. versionadded:: 20.0
         """
         return _SocketContext(self, 'connect', addr)
 
-    def _bind_cm(self: T, addr: str) -> _SocketContext[T]:
+    def _bind_cm(self: _SocketType, addr: str) -> _SocketContext[_SocketType]:
         """Context manager to unbind on exit
 
         .. versionadded:: 20.0
         """
+        try:
+            # retrieve last_endpoint
+            # to support binding on random ports via
+            # `socket.bind('tcp://127.0.0.1:0')`
+            addr = cast(bytes, self.get(zmq.LAST_ENDPOINT)).decode("utf8")
+        except (AttributeError, ZMQError, UnicodeDecodeError):
+            pass
         return _SocketContext(self, 'bind', addr)
 
-    def bind(self: T, addr: str) -> _SocketContext[T]:
+    def bind(self: _SocketType, addr: str) -> _SocketContext[_SocketType]:
         """s.bind(addr)
 
         Bind the socket to an address.
@@ -289,6 +299,11 @@ class Socket(SocketBase, AttributeSetter, Generic[ST]):
         .. versionadded:: 20.0
             Can be used as a context manager.
 
+        .. versionadded:: 26.0
+            binding to port 0 can be used as a context manager
+            for binding to a random port.
+            The URL can be retrieved as `socket.last_endpoint`.
+
         Parameters
         ----------
         addr : str
@@ -305,7 +320,7 @@ class Socket(SocketBase, AttributeSetter, Generic[ST]):
             raise
         return self._bind_cm(addr)
 
-    def connect(self: T, addr: str) -> _SocketContext[T]:
+    def connect(self: _SocketType, addr: str) -> _SocketContext[_SocketType]:
         """s.connect(addr)
 
         Connect to a remote 0MQ socket.
@@ -382,7 +397,7 @@ class Socket(SocketBase, AttributeSetter, Generic[ST]):
         """
         return self.FD
 
-    def subscribe(self, topic: Union[str, bytes]) -> None:
+    def subscribe(self, topic: str | bytes) -> None:
         """Subscribe to a topic
 
         Only for SUB sockets.
@@ -393,7 +408,7 @@ class Socket(SocketBase, AttributeSetter, Generic[ST]):
             topic = topic.encode('utf8')
         self.set(zmq.SUBSCRIBE, topic)
 
-    def unsubscribe(self, topic: Union[str, bytes]) -> None:
+    def unsubscribe(self, topic: str | bytes) -> None:
         """Unsubscribe from a topic
 
         Only for SUB sockets.
@@ -449,7 +464,7 @@ class Socket(SocketBase, AttributeSetter, Generic[ST]):
     getsockopt_unicode = getsockopt_string = get_string
 
     def bind_to_random_port(
-        self: T,
+        self: _SocketType,
         addr: str,
         min_port: int = 49152,
         max_port: int = 65536,
@@ -487,7 +502,7 @@ class Socket(SocketBase, AttributeSetter, Generic[ST]):
         ):
             # if LAST_ENDPOINT is supported, and min_port / max_port weren't specified,
             # we can bind to port 0 and let the OS do the work
-            self.bind("%s:*" % addr)
+            self.bind(f"{addr}:*")
             url = cast(bytes, self.last_endpoint).decode('ascii', 'replace')
             _, port_s = url.rsplit(':', 1)
             return int(port_s)
@@ -576,10 +591,9 @@ class Socket(SocketBase, AttributeSetter, Generic[ST]):
         copy: bool = ...,
         *,
         track: Literal[True],
-        routing_id: Optional[int] = ...,
-        group: Optional[str] = ...,
-    ) -> "zmq.MessageTracker":
-        ...
+        routing_id: int | None = ...,
+        group: str | None = ...,
+    ) -> zmq.MessageTracker: ...
 
     @overload
     def send(
@@ -589,10 +603,9 @@ class Socket(SocketBase, AttributeSetter, Generic[ST]):
         copy: bool = ...,
         *,
         track: Literal[False],
-        routing_id: Optional[int] = ...,
-        group: Optional[str] = ...,
-    ) -> None:
-        ...
+        routing_id: int | None = ...,
+        group: str | None = ...,
+    ) -> None: ...
 
     @overload
     def send(
@@ -601,10 +614,9 @@ class Socket(SocketBase, AttributeSetter, Generic[ST]):
         flags: int = ...,
         *,
         copy: bool = ...,
-        routing_id: Optional[int] = ...,
-        group: Optional[str] = ...,
-    ) -> None:
-        ...
+        routing_id: int | None = ...,
+        group: str | None = ...,
+    ) -> None: ...
 
     @overload
     def send(
@@ -613,10 +625,9 @@ class Socket(SocketBase, AttributeSetter, Generic[ST]):
         flags: int = ...,
         copy: bool = ...,
         track: bool = ...,
-        routing_id: Optional[int] = ...,
-        group: Optional[str] = ...,
-    ) -> Optional["zmq.MessageTracker"]:
-        ...
+        routing_id: int | None = ...,
+        group: str | None = ...,
+    ) -> zmq.MessageTracker | None: ...
 
     def send(
         self,
@@ -624,9 +635,9 @@ class Socket(SocketBase, AttributeSetter, Generic[ST]):
         flags: int = 0,
         copy: bool = True,
         track: bool = False,
-        routing_id: Optional[int] = None,
-        group: Optional[str] = None,
-    ) -> Optional["zmq.MessageTracker"]:
+        routing_id: int | None = None,
+        group: str | None = None,
+    ) -> zmq.MessageTracker | None:
         """Send a single zmq message frame on this socket.
 
         This queues the message to be sent by the IO thread at a later time.
@@ -657,8 +668,8 @@ class Socket(SocketBase, AttributeSetter, Generic[ST]):
         None : if `copy` or not track
             None if message was sent, raises an exception otherwise.
         MessageTracker : if track and not copy
-            a MessageTracker object, whose `pending` property will
-            be True until the send is completed.
+            a MessageTracker object, whose `done` property will
+            be False until the send is completed.
 
         Raises
         ------
@@ -727,8 +738,8 @@ class Socket(SocketBase, AttributeSetter, Generic[ST]):
         -------
         None : if copy or not track
         MessageTracker : if track and not copy
-            a MessageTracker object, whose `pending` property will
-            be True until the last send is completed.
+            a MessageTracker object, whose `done` property will
+            be False until the last send is completed.
         """
         # typecheck parts before sending:
         for i, msg in enumerate(msg_parts):
@@ -741,11 +752,7 @@ class Socket(SocketBase, AttributeSetter, Generic[ST]):
                 if len(rmsg) > 32:
                     rmsg = rmsg[:32] + '...'
                 raise TypeError(
-                    "Frame %i (%s) does not support the buffer interface."
-                    % (
-                        i,
-                        rmsg,
-                    )
+                    f"Frame {i} ({rmsg}) does not support the buffer interface."
                 )
         for msg in msg_parts[:-1]:
             self.send(msg, zmq.SNDMORE | flags, copy=copy, track=track)
@@ -755,28 +762,24 @@ class Socket(SocketBase, AttributeSetter, Generic[ST]):
     @overload
     def recv_multipart(
         self, flags: int = ..., *, copy: Literal[True], track: bool = ...
-    ) -> List[bytes]:
-        ...
+    ) -> list[bytes]: ...
 
     @overload
     def recv_multipart(
         self, flags: int = ..., *, copy: Literal[False], track: bool = ...
-    ) -> List[zmq.Frame]:
-        ...
+    ) -> list[zmq.Frame]: ...
 
     @overload
-    def recv_multipart(self, flags: int = ..., *, track: bool = ...) -> List[bytes]:
-        ...
+    def recv_multipart(self, flags: int = ..., *, track: bool = ...) -> list[bytes]: ...
 
     @overload
     def recv_multipart(
         self, flags: int = 0, copy: bool = True, track: bool = False
-    ) -> Union[List[zmq.Frame], List[bytes]]:
-        ...
+    ) -> list[zmq.Frame] | list[bytes]: ...
 
     def recv_multipart(
         self, flags: int = 0, copy: bool = True, track: bool = False
-    ) -> Union[List[zmq.Frame], List[bytes]]:
+    ) -> list[zmq.Frame] | list[bytes]:
         """Receive a multipart message as a list of bytes or Frame objects
 
         Parameters
@@ -890,7 +893,7 @@ class Socket(SocketBase, AttributeSetter, Generic[ST]):
         copy: bool = True,
         encoding: str = 'utf-8',
         **kwargs,
-    ) -> Optional["zmq.Frame"]:
+    ) -> zmq.Frame | None:
         """Send a Python unicode string as a message with an encoding.
 
         0MQ communicates with raw bytes, so you must encode/decode
@@ -902,7 +905,7 @@ class Socket(SocketBase, AttributeSetter, Generic[ST]):
             The unicode string to send.
         flags : int, optional
             Any valid flags for :func:`Socket.send`.
-        encoding : str [default: 'utf-8']
+        encoding : str
             The encoding to be used
         """
         if not isinstance(u, str):
@@ -918,7 +921,7 @@ class Socket(SocketBase, AttributeSetter, Generic[ST]):
         ----------
         flags : int
             Any valid flags for :func:`Socket.recv`.
-        encoding : str [default: 'utf-8']
+        encoding : str
             The encoding to be used
 
         Returns
@@ -929,7 +932,7 @@ class Socket(SocketBase, AttributeSetter, Generic[ST]):
         Raises
         ------
         ZMQError
-            for any of the reasons :func:`~Socket.recv` might fail
+            for any of the reasons :func:`Socket.recv` might fail
         """
         msg = self.recv(flags=flags)
         return self._deserialize(msg, lambda buf: buf.decode(encoding))
@@ -938,8 +941,18 @@ class Socket(SocketBase, AttributeSetter, Generic[ST]):
 
     def send_pyobj(
         self, obj: Any, flags: int = 0, protocol: int = DEFAULT_PROTOCOL, **kwargs
-    ) -> Optional[zmq.Frame]:
-        """Send a Python object as a message using pickle to serialize.
+    ) -> zmq.Frame | None:
+        """
+        Send a Python object as a message using pickle to serialize.
+
+        .. warning::
+
+            Never deserialize an untrusted message with pickle,
+            which can involve arbitrary code execution.
+            Make sure to authenticate the sources of messages
+            before unpickling them, e.g. with transport-level security
+            (e.g. CURVE, ZAP, or IPC permissions)
+            or signed messages.
 
         Parameters
         ----------
@@ -955,7 +968,17 @@ class Socket(SocketBase, AttributeSetter, Generic[ST]):
         return self.send(msg, flags=flags, **kwargs)
 
     def recv_pyobj(self, flags: int = 0) -> Any:
-        """Receive a Python object as a message using pickle to serialize.
+        """
+        Receive a Python object as a message using UNSAFE pickle to serialize.
+
+        .. warning::
+
+            Never deserialize an untrusted message with pickle,
+            which can involve arbitrary code execution.
+            Make sure to authenticate the sources of messages
+            before unpickling them, e.g. with transport-level security
+            (such as CURVE or IPC permissions)
+            or authenticating messages themselves before deserializing.
 
         Parameters
         ----------
@@ -994,7 +1017,7 @@ class Socket(SocketBase, AttributeSetter, Generic[ST]):
         msg = jsonapi.dumps(obj, **kwargs)
         return self.send(msg, flags=flags, **send_kwargs)
 
-    def recv_json(self, flags: int = 0, **kwargs) -> Union[List, str, int, float, Dict]:
+    def recv_json(self, flags: int = 0, **kwargs) -> list | str | int | float | dict:
         """Receive a Python object as a message using json to serialize.
 
         Keyword arguments are passed on to json.loads
@@ -1019,16 +1042,18 @@ class Socket(SocketBase, AttributeSetter, Generic[ST]):
 
     _poller_class = Poller
 
-    def poll(self, timeout=None, flags=zmq.POLLIN) -> int:
+    def poll(self, timeout: int | None = None, flags: int = zmq.POLLIN) -> int:
         """Poll the socket for events.
+
         See :class:`Poller` to wait for multiple sockets at once.
 
         Parameters
         ----------
-        timeout : int [default: None]
+        timeout : int
             The timeout (in milliseconds) to wait for an event. If unspecified
             (or specified None), will wait forever for an event.
-        flags : int [default: POLLIN]
+        flags : int
+            default: POLLIN.
             POLLIN, POLLOUT, or POLLIN|POLLOUT. The event flags to poll for.
 
         Returns
@@ -1048,8 +1073,8 @@ class Socket(SocketBase, AttributeSetter, Generic[ST]):
         return evts.get(self, 0)
 
     def get_monitor_socket(
-        self: T, events: Optional[int] = None, addr: Optional[str] = None
-    ) -> T:
+        self: _SocketType, events: int | None = None, addr: str | None = None
+    ) -> _SocketType:
         """Return a connected PAIR socket ready to receive the event notifications.
 
         .. versionadded:: libzmq-4.0
@@ -1057,20 +1082,21 @@ class Socket(SocketBase, AttributeSetter, Generic[ST]):
 
         Parameters
         ----------
-        events : int [default: ZMQ_EVENT_ALL]
+        events : int
+            default: `zmq.EVENT_ALL`
             The bitmask defining which events are wanted.
-        addr :  string [default: None]
+        addr : str
             The optional endpoint for the monitoring sockets.
 
         Returns
         -------
-        socket :  (PAIR)
-            The socket is already connected and ready to receive messages.
+        socket : zmq.Socket
+            The PAIR socket, connected and ready to receive messages.
         """
         # safe-guard, method only available on libzmq >= 4
         if zmq.zmq_version_info() < (4,):
             raise NotImplementedError(
-                "get_monitor_socket requires libzmq >= 4, have %s" % zmq.zmq_version()
+                f"get_monitor_socket requires libzmq >= 4, have {zmq.zmq_version()}"
             )
 
         # if already monitoring, return existing socket
@@ -1103,4 +1129,6 @@ class Socket(SocketBase, AttributeSetter, Generic[ST]):
         self.monitor(None, 0)
 
 
-__all__ = ['Socket']
+SyncSocket: TypeAlias = Socket[bytes]
+
+__all__ = ['Socket', 'SyncSocket']
