@@ -2,7 +2,7 @@
 //  Copyright (c) 2010 Brian Granger, Min Ragan-Kelley
 //
 //  Distributed under the terms of the New BSD License.  The full license is in
-//  the file COPYING.BSD, distributed as part of this software.
+//  the file LICENSE.BSD, distributed as part of this software.
 //-----------------------------------------------------------------------------
 
 #pragma once
@@ -18,8 +18,6 @@
 
 
 #include "zmq.h"
-// version compatibility for constants:
-#include "zmq_constants.h"
 
 #define _missing (-1)
 
